package net.pingex.dcf.commands;

import net.pingex.dcf.commands.audit.AuditManager;
import net.pingex.dcf.commands.audit.AuditResult;
import net.pingex.dcf.commands.parser.BasicParser;
import net.pingex.dcf.commands.parser.ICommandParser;
import net.pingex.dcf.commands.parser.ParserException;
import net.pingex.dcf.core.Configuration;
import net.pingex.dcf.modularity.PluginWrapper;
import net.pingex.dcf.permissions.PermissionsHandler;
import net.pingex.dcf.util.DiscordInteractionsUtil;
import org.apache.commons.lang3.concurrent.BasicThreadFactory;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import sx.blah.discord.handle.impl.events.MessageReceivedEvent;
import java.lang.reflect.Method;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/**
 * Landing class for the whole command package
 */
public class CommandHandler
{
	/**
	 * The one and only valid parser for now.
	 */
	private static final ICommandParser PARSER = new BasicParser();

	/**
	 * Logger
	 */
	private static final Logger LOGGER = LogManager.getLogger(CommandHandler.class);

	/**
	 * Executor Service for dispatching commands to threads
	 * TODO: ExecutorService.shutdown() when using a stop command or whatever
	 */
	private static final ExecutorService threadPool = Executors.newCachedThreadPool(new BasicThreadFactory.Builder().namingPattern("CommandHandler-%d").build());

	/**
	 * Landing method for a MRE
	 */
	public static void handle(MessageReceivedEvent event)
	{
		String command;
		List<String> arguments;

		// Parse
		try
		{
			command = PARSER.parseCommand(event.getMessage().getContent());
			arguments = PARSER.parseArguments(event.getMessage().getContent());
		}
		catch(ParserException e)
		{
			return; // Not a command
		}
		LOGGER.debug("Attempting to run command {} for user #{}.", command, event.getMessage().getAuthor().getID());

		// Query command from the bank
		Optional<Command> targetCommand = CommandRegistry.getCommandOrAliasByName(command);
		if(!targetCommand.isPresent())
		{
			LOGGER.debug("No command with that name was found.");
			DiscordInteractionsUtil.sendMessage(event.getMessage().getChannel(), "Command not found. Use `"+ Configuration.COMMAND_PREFIX + "internal:list` to get a list of all available commands.");
			return;
		}

		// Audit
		AuditResult result = AuditManager.doAudit(new Context(targetCommand.get(), arguments, event.getMessage(), event.getClient()));
		if(result.getOpcode().equals(AuditResult.ResultCode.FAIL))
		{
			LOGGER.info("Denied command {} for user #{}. OPCode: {},  Reason: {}", targetCommand.get().getName(), event.getMessage().getAuthor().getID(), result.getOpcode(), result.getMessage());
			DiscordInteractionsUtil.sendMessage(event.getMessage().getChannel(), result.getMessage());
			return;
		}
		if(result.getOpcode().equals(AuditResult.ResultCode.WARN))
			DiscordInteractionsUtil.sendMessage(event.getMessage().getChannel(), "Warning: " + result.getMessage());

		LOGGER.info("Accepted command {} for user #{}. OPCode: {},  Reason: {}", targetCommand.get().getName(), event.getMessage().getAuthor().getID(), result.getOpcode(), result.getMessage() != null ? result.getMessage() : "N/A");

		// Run it
		threadPool.submit(() ->
		{
			try
			{
				targetCommand.get().execute(event, arguments);
			}
			catch(Throwable throwable)
			{
				LOGGER.error("Error while executing command " + command + ".", throwable);
				DiscordInteractionsUtil.sendMessage(event.getMessage().getChannel(), "Error while executing the command.");
			}
		});
	}

	/**
	 * Registers a plugin's commands, if any
	 * @param pluginWrapper Target plugin
	 */
	public static void registerPlugin(PluginWrapper pluginWrapper)
	{
		LOGGER.debug("Registering commands for plugin {}.", pluginWrapper.getId());

		// Raw commands
		if(IWithCommands.class.isAssignableFrom(pluginWrapper.getInstance().getClass()))
		{
			LOGGER.debug("Plugin has commands (raw commands).");
			((IWithCommands) pluginWrapper.getInstance()).getCommands().forEach(CommandRegistry::registerCommand);
		}

		// Annotated commands
		if(IWithAnnotatedCommands.class.isAssignableFrom(pluginWrapper.getInstance().getClass()))
		{
			LOGGER.debug("Plugin has commands (annotated commands).");
			((IWithAnnotatedCommands) pluginWrapper.getInstance()).getAnnotatedObjects().forEach(e ->
			{
				for(Method i : e.getClass().getMethods())
					if(i.isAnnotationPresent(AnnotatedCommand.class))
						CommandRegistry.registerCommand(Command.buildFromAnnotatedCommand(i.getAnnotation(AnnotatedCommand.class), i, e));
			});
		}
	}

	/**
	 * Unregisters a plugin's commands, if any
	 * @param pluginWrapper Target plugin
	 */
	public static void unregisterPlugin(PluginWrapper pluginWrapper)
	{
		LOGGER.debug("Removing commands for plugin {}.", pluginWrapper.getId());

		// Raw commands
		if(IWithCommands.class.isAssignableFrom(pluginWrapper.getInstance().getClass()))
		{
			LOGGER.debug("Plugin has commands (raw commands).");
			((IWithCommands) pluginWrapper.getInstance()).getCommands().forEach(CommandRegistry::unregisterCommand);
		}

		// Annotated commands
		if(IWithAnnotatedCommands.class.isAssignableFrom(pluginWrapper.getInstance().getClass()))
		{
			LOGGER.debug("Plugin has commands (annotated commands).");
			((IWithAnnotatedCommands) pluginWrapper.getInstance()).getAnnotatedObjects().forEach(e ->
			{
				for(Method i : e.getClass().getMethods())
					if(i.isAnnotationPresent(AnnotatedCommand.class))
						CommandRegistry.unregisterCommandByName(i.getAnnotation(AnnotatedCommand.class).name());
			});
		}
	}
}
